#!/usr/bin/python3


# This is a sample Python script.

# Press Shift+F10 to execute it or replace it with your code.
# Press Double Shift to search everywhere for classes, files, tool windows, actions, and settings.

import sys
from paftv.lib import alitv, filter
#from collections import defaultdict
#import itertools
#import json
#import logging
from enum import Enum
from pathlib import Path
from Bio import SeqIO
import argparse


"""
TODO: 
- Option to automatic translocations detection? 


"""



class Alignment:
    """
    A single alignment region between two genomes.
    This class implements values required to be converted to AliTV Link objects.
    """

    def __init__(self, query_name, query_start, query_end,
                 target_name, target_start, target_end,
                 identity, strand, length):
        """
        :param query_name: Query sequence name
        :type query_name: str
        :param query_start: Query start (0-based)
        :type query_start: int
        :param query_end: Query end (0-based)
        :type query_end: int
        :param target_name: Target sequence name
        :type target_name: str
        :param target_start: Target start (0-based)
        :type target_start: int
        :param target_end: Target end (0-based)
        :type target_end: int
        :param identity: Percent identity of a sequence, represented as a float between 0 and 100.
        :type identity: float
        :param strand: Relative strand
        :type strand: Strand
        :param length: Alignment length
        :type length: int
        """

        self.query_name = query_name
        self.query_start = query_start
        self.query_end = query_end

        self.target_name = target_name
        self.target_start = target_start
        self.target_end = target_end


        self.identity = identity
        if isinstance(strand, Strand):
            self.strand = strand
        else:
            raise ValueError(
                "{} is not a valid Alignment.strand value. Alignment objects need to be instantiated with a member of"
                " minitv.align.Strand.".format(type(strand)))
        self.length = length

    def __len__(self):
        return self.length

class PAFAlignment(Alignment):
    """
    A single alignment between two genomes in PAF format.
    """

    def __init__(self, query_name, query_len, query_start, query_end, strand, target_name, target_len, target_start,
                 target_end, matches, length, mapping_quality, id, tags):
        """
        :param query_name: Query sequence name
        :type query_name: str
        :param query_len: Query sequence length
        :type query_len: int
        :param query_start: Query start (0-based)
        :type query_start: int
        :param query_end: Query end (0-based)
        :type query_end: int
        :param strand: Relative strand: "+" or "-"
        :type strand: str
        :param target_name: Target sequence name
        :type target_name: str
        :param target_len: Target sequence length
        :type target_len: int
        :param target_start: Target start on original strand (0-based)
        :type target_start: int
        :param target_end: Target end on original strand (0-based)
        :type target_end: int
        :param matches: Number of residue matches
        :type matches: int
        :param length: Alignment block length
        :type length: int
        :param mapping_quality: Mapping quality (0-255; 255 for missing)
        :type mapping_quality: int
        :param tags: A dictionary of SAM-like typed key-value pairs
        :type tags: dict
        """

        if strand == '+':
            strand = Strand.FORWARD
        elif strand == '-':
            strand = Strand.REVERSE
        else:
            raise ValueError("{} is an invalid PAF alignment strand value.".format(strand))


        ## This is the identity
        if id != 0:
            identity = id
        else:
            identity = (matches / length) * 100
        super().__init__(query_name, query_start, query_end, target_name, target_start, target_end, identity,
                         strand, length)

        self.query_len = query_len
        self.target_len = target_len
        self.matches = matches

        if mapping_quality < 0 or mapping_quality > 255:
            raise ValueError("'{}' is an invalid mapping quality for the PAF format. Mapping quality must be between 0 "
                             "and 255.".format(mapping_quality))
        else:
            self.mapping_quality = mapping_quality

        self._tags = tags

    def get_tag(self, key):
        return self._tags[key]

    def set_tag(self, key, value):
        self._tags[key] = value


    # Compare
    def __eq__(self, other):
        if not isinstance(other, PAFAlignment):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return self.query_end == other.query_end and \
               self.query_start == other.query_start and \
                self.target_start == other.target_start and \
                self.target_end == other.target_end and \
                self.target_name == other.target_name and \
                self.query_name == other.query_name and \
                self.target_name == other.target_name


def parse_output_minimap2(output, id):
    """
    :param output: minimap2 stdout
    :type output: io.TextIOBase or io.StringIO
    :rtype: list[Alignment]
    """

    alignments = []
    for line in output:

        split_line = line.rstrip().split("\t")
        query_name = split_line[0]
        query_len = int(split_line[1])
        query_start = int(split_line[2])
        query_end = int(split_line[3])
        strand = split_line[4]
        target_name = split_line[5]
        target_len = int(split_line[6])
        target_start = int(split_line[7])
        target_end = int(split_line[8])
        matches = int(split_line[9])
        length = int(split_line[10])
        mapping_quality = int(split_line[11])
        id_numb = 0
        if id:
            if len(split_line) > 12 and len(set([x for x in split_line[12:] if x.startswith("id")])):
                id_numb = float(split_line[12].split(":")[-1])
            else:
                print("Ignoring id tag - not existing")
                id = False
        tags = {}

        if len(split_line) > 12:
            # parse SAM tags
            key_value_pairs = split_line[12:]
            for key_value_pair in key_value_pairs:
                key, type_char, value = key_value_pair.split(':')
                if type_char == "i":
                    tags[key] = int(value)
                elif type_char == 'f':
                    tags[key] = float(value)
                else:
                    tags[key] = value

        new_alignment = PAFAlignment(query_name, query_len, query_start, query_end, strand, target_name,
                                     target_len, target_start, target_end, matches, length, mapping_quality, id_numb, tags)
        alignments.append(new_alignment)

    return alignments

def parseGroups(group_file, xornot):
    """

    :param group_file:

    name, fastaid1, fastfa2
    :return:
    """

    data = dict()
    count = 0
    with open(group_file) as f:
        for lines in f.readlines():
            lsplit = [x.replace("\n","").replace(" ", "") for x in lines.split(",")]
            if xornot:
                p = [x + "_1" for x in lsplit]
                data[count] = lsplit + p
            else:
                data[count] = lsplit
            count += 1
    o = dict()
    for k,v in data.items():
        for x in v:
            o[x] = k
    return o, data



class FASTAFile:
    """
    Represents a FASTA file of one or more sequences.
    """

    def __init__(self, path):
        """
        :param path: path to the FASTA file
        :type path: pathlib.Path
        """

        self.path = path
        self.name = path.stem

        self.records = list(SeqIO.parse(str(path), 'fasta'))

    def __len__(self):
        """
        :return: number of records in the FASTA file
        """

        return len(self.records)


class Strand(Enum):
    """ File schema independent method to encode alignment strand. """
    FORWARD = 1
    UNKNOWN = 0
    REVERSE = -1

# Press the green button in the gutter to run the script.
if __name__ == '__main__':

    main()






def main():

    ## ARGUMENT PARSING

    parser = argparse.ArgumentParser();
    parser.add_argument(
        "-p",
        "--paf",
        metavar="paf",
        help="Alignment file in paf format",
        required=True
    )

    parser.add_argument(
        "-q",
        "--query",
        metavar="query.fa",
        nargs='+',
        help="FASTA file of sequence(s). This argument can instead take a single file of file names containing "
             "paths relative to the working directory.",
        required=True
    )

    # parser.add_argument(
    #     "-r",
    #     "--region",
    #     help="Bounds of region to view, in standard samtools region format (refSeqID:start-end)"
    # )
    #
    # parser.add_argument(
    #     "--min_link_identity",
    #     help="Minimum link identity for AliTV to initially display",
    #     type=float,
    #     default=0.0
    # )
    #
    # parser.add_argument(
    #     "--min_link_length",
    #     help="Minimum link length for AliTV to initially display",
    #     type=int,
    #     default=0
    # )
    #
    # parser.add_argument(
    #     "--min_aln_cov",
    #     help="Minimum percent coverage, from 0 to 100, by alignments that a sequence needs to be initially visible in "
    #          "AliTV.",
    #     type=float,
    #     default=0.0
    # )
    #
    # parser.add_argument(
    #     "--min_ref_cov",
    #     help="Minimum percent coverage, from 0 to 100, of a query sequence by alignments to the reference needed for"
    #          "the query sequence to be initially visible in AliTV.",
    #     type=float,
    #     default=0.0
    # )

    parser.add_argument(
        "-o",
        "--output",
        help="Output name",
        required=True,
    )

    parser.add_argument(
        "-X",
        "--allvsall",
        help="Query and the reference are the same (all-vs-all alignment) - only one genome [default: off]",
        # required=True,
        action="store_true"

    )
    parser.add_argument(
        "-c",
        "--color",
        help="Min, mid and max identity colors in hex format (without #) example 'D21414,FFEE05,1DAD0A (also possible in AliTV)",
        type=str
    )

    # parser.add_argument(
    #     "-t",
    #     "--transposon",
    #     help = "Only show 'jumping' alignments",
    #     action="store_true"
    # )

    # parser.add_argument(
    #     "--maxiteration",
    #     help = "Number of maximal iteration if region linking is used (only works with -r, ignored otherwise)",
    #     type = int
    # )

    # parser.add_argument(
    #     "-g",
    #     "--group_list",
    #     help = "CSV format - each 'group' (e.g. same chromosome) in one line. fasta_entry1, fasta_entry2, fasta_entry3 --> Check examples"
    # )

    parser.add_argument(
        "--mmid",
        help="Minimum and maximum identity range (e.g. 0,100) [default: dynamic min/max of all alignments]"
    )

    args = parser.parse_args()

    # Make a fasta file list
    paths = tuple([Path(x) for x in args.query])
    fasta_files = [FASTAFile(x) for x in paths]

    alignment_groups = {}  # key = Tuple(target, query)
    alignment_lists = {}  # This one saves the alignments in the same style as alignment groups, but holds a list of strings (lines) as values
    paf = args.paf

    if args.allvsall:
        if len(fasta_files) != 1:
            print("Only provide one fasta file with the -X flag")
            sys.exit()
        else:
            p = FASTAFile(Path(args.query[0]))
            p.name = p.name + "_1"
            for x in p.records:
                x.name = x.name + "_1"
                x.id = x.id + "_1"
            fasta_files.append(p)

            with open(paf) as f:

                for line1 in f.readlines():
                    split_line = line1.rstrip().split("\t")
                    split_line[0] = split_line[0] + "_1"
                    line2 = "\t".join(split_line)
                    # print(alignment_lists.get((split_line[5], split_line[0])))
                    if alignment_lists.get((fasta_files[0].name, fasta_files[1].name)) != None:
                        # print(line1.rstrip(),
                        alignment_lists[fasta_files[0].name, fasta_files[1].name].append(line2.rstrip())
                    else:
                        # print(len(line1))
                        alignment_lists[fasta_files[0].name, fasta_files[1].name] = [line2.rstrip()]
            for k, v in alignment_lists.items():
                alignment_groups[k] = parse_output_minimap2(v, False)



    # If you have two distict input files
    else:
        '''
        # This is to find the right FASTA file afterwards
        Dict[entry] --> Fasta file

        '''

        entry2fasta = dict()
        entrycheck = []
        for x in fasta_files:
            for y in x.records:
                entrycheck.append(x)
                entry2fasta[y.id] = x.name

        # Check if we have double entry -> I
        if len(entrycheck) != len(set(entrycheck)):
            print("Duplicated fasta entries")

        all_combinations = set()
        with open(paf) as f:

            for line1 in f.readlines():
                split_line = line1.rstrip().split("\t")
                # print(alignment_lists.get((split_line[5], split_line[0])))
                if entry2fasta.get(split_line[0]) != None:
                    if entry2fasta.get(split_line[5]) != None:
                        if entry2fasta[split_line[0]] != entry2fasta[split_line[5]]:

                            query = entry2fasta[split_line[0]]
                            target = entry2fasta[split_line[5]]

                            all_combinations.add((target, query))

                            # Sort out self alignments
                            if query == target:
                                continue;
                            else:
                                if alignment_lists.get((target, query)) != None:
                                    alignment_lists[(target, query)].append(line1.rstrip())
                                else:
                                    alignment_lists[(target, query)] = [line1.rstrip()]

        for k, v in alignment_lists.items():
            alignment_groups[k] = parse_output_minimap2(v, False)

    filter.reduceAlg(alignment_groups)
    ## If you define a region, make a new alignment
    new_alg = {}
    count = 0

    # if args.region != None:
    #     region = [[args.region.split(":")[-2], int(args.region.split(":")[-1].split("-")[0]),
    #                int(args.region.split(":")[-1].split("-")[1])]]
    #     if args.maxiteration == None:
    #         new_alg = filter.filter_region(alignment_groups, region)
    #     else:
    #         new_alg = filter.filter_region(alignment_groups, region, args.maxiteration)
    #     alignment_groups = new_alg

    # if args.group_list != None:
    #     new_alg = {}
    #     data, gen = parseGroups(args.group_list, args.allvsall)
    #     new_alg = filter.filter_transpon(gen, data, alignment_groups)
    #     alignment_groups = new_alg

    # We calculate the max and min identity
    minId = 100
    maxId = 0
    if args.mmid != None:
        minId = int(args.mmid.split(",")[0])
        maxId = int(args.mmid.split(",")[1])
    else:
        for k, v in alignment_groups.items():
            for x in v:
                if minId > x.identity:
                    minId = x.identity
                if maxId < x.identity:
                    maxId = x.identity

    # -----------------------------------------------------------------------------------------------------------------------------------------------------
    # -----------------------------------------------------------------------------------------------------------------------------------------------------

    # Try new functions here:
    # filter.overlaps(alignment_groups)

    """ 
    This is Alitv - try not to touch this
    Changes:
        - min, max ID in AliTv (maybe change to function)
        - changeColors function 
        - If regions given - use new alignment group  #
        - Removed Order and orient sequence 
    """
    ali_tv = alitv.AliTV(fasta_files, minId, maxId)

    if args.color != None:
        ali_tv.changeColors(args.color.split(","))

    ali_tv.load_links(alignment_groups)

    # ali_tv.set_soft_filters(args.min_link_identity, args.min_link_length)

    # if args.min_aln_cov:
    #     print('Hiding chromosomes by alignment coverage')
    #     ali_tv.hide_chromosomes_by_alignment_coverage(args.min_aln_cov)

    # if args.min_ref_cov:
    #     print('Hiding chromosomes by reference alignment coverage')
    #     ali_tv.hide_chromosomes_by_reference_coverage(args.min_ref_cov)

    # ali_tv.order_and_orient_sequences(alignment_groups)

    ali_tv.optimize_configuration()

    # logging.info('Generating JSON output')

    # Leave it like it is
    a = ali_tv.get_json(indent=1)

    with open(args.output, "w") as l:
        l.write(a)

# See PyCharm help at https://www.jetbrains.com/help/pycharm/



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 2/2/21

@author: moinSebi

"""


def filter_region(alg_groups: dict(), region: list, maxit: int = 1000) -> dict():
    """
    :param alg_groups:
    :param region:
    :return:
    """
    r = region
    new_alg = dict()
    count = 0

    delete = []
    while (True):
        nothing_new = True
        p = []
        for key, value in alg_groups.items():
            if new_alg.get(key) == None:
                new_alg[key] = []
            for alignment in value:
                # print(type(x))
                for entry in region:
                    if entry[0] == alignment.query_name:
                        # print("hit1")
                        if getOverlap([alignment.query_start, alignment.query_end], entry):
                            j = True
                            for d in delete:
                                if d == alignment:
                                    j = False
                            if j:
                                p.append([alignment.target_name, alignment.target_start, alignment.target_end])
                                new_alg[key].append(alignment)
                                nothing_new = False
                                delete.append(alignment)


                    elif entry[0] == alignment.target_name:
                        # print("hit2")
                        if getOverlap([alignment.target_start, alignment.target_end], entry):
                            #print(alignment)
                            j = True
                            for d in delete:
                                if d == alignment:
                                    j = False
                            if j:
                                nothing_new = False
                                p.append([alignment.query_name, alignment.query_start, alignment.query_end])
                                new_alg[key].append(alignment)
                                delete.append(alignment)
        print("Iteration: {}".format(count))
        print("New Alignments: {}".format(len(p)))
        print("Total Alignments: {}".format(len(delete)))
        print()
        region = []
        region.extend(p)
        count += 1
        if nothing_new or count == maxit+1:
            break


    return new_alg

def reduceAlg(alg_groups: dict) -> dict:

    new_alg_temp = dict()

    for k, v in alg_groups.items():
        new_alg_temp[k] = sorted(v, key = lambda x: (x.query_name, x.target_name, x.target_start, x.target_end, x.query_start, x.query_end))






def getOverlap(start_end: list, region: list ) -> bool:
    return max(0, min(start_end[1], region[2]) - max(start_end[0], region[1]))

def filter_transpon(t_list: list, gen_dict: dict, alg_groups: dict) -> dict:

    new_alg = dict()

    for k, v in alg_groups.items():
        new_alg[k] = []
        for alignment in v:
            if (alignment.query_name == "TAIR10_Chr1_1"):
                #print("hjdahskjda")
                pass
            if alignment.target_name in list(gen_dict.keys()) and alignment.query_name in list(gen_dict.keys()):
                #print(alignment.target_name)
                if alignment.target_name not in t_list[gen_dict[alignment.query_name]]:
                    new_alg[k].append(alignment)
            else:
                new_alg[k].append(alignment)

    return new_alg


def getOverlap1(start_end: list, region: list ) -> bool:
    return max(0, min(start_end[1], region[1]) - max(start_end[0], region[0]))

def overlaps(alg_dict: dict()):
    """

    :param alg_dict:
    :return:
    """

    
    """"
    TODO:
     - This will only work with 2 genomes
     - Care of k in alg_dict
     - Maybe also to able to give list


     """

    # if overlap
    check = dict()

    for k,v in alg_dict.items():
        for x in v:
            if check.get((x.query_name, x.target_name)) == None:
                check[(x.query_name, x.target_name)]  = [[x.target_start, x.target_end]]
            else:
                check[(x.query_name, x.target_name)].append([x.target_start, x.target_end])

    for k, v in check.items():
        #print(k)
        #print(v)
        v2 = sorted(v, key = lambda x: x[0])
        check[k] = v2

    odict = dict()
    for k,v in check.items():
            print(len(v))
            print(v)
            overlapList = []
            for index in range(len(v)):
                if len(overlapList) > 0:
                    if getOverlap1(overlapList[-1], v[index]):
                        overlapList[-1] = [min(overlapList[-1][0], v[index][0]), max(overlapList[-1][1], v[index][1])]
                    else:
                        overlapList.append(v[index])
                else:
                    overlapList.append(v[index])
            odict[k] = overlapList

    sdict = dict()
    for k, v in odict.items():
        p = 0
        for x in v:
            p += x[1] -x[0]
        sdict[k] = p
        print(p)

    allfiles = set()
    for k, v in sdict.items():
        allfiles.add(k[0])
    for k, v in sdict.items():
        if k[0].startswith("10015"):
            print(k)
            print(v)

    maxes = dict()
    for x in allfiles:
        op = []
        for k,v in sdict.items():
            if k[0] == x:
                op.append((k, v))
        op = sorted(op, key = lambda d: d[1])
        maxes[x] = (op[-1][0][1], op[-1][1])
    #print(allfiles)














#!/usr/bin/python3

"""
Handles AliTV integration
"""

from collections import defaultdict
#import dendropy
#import itertools
import json
from enum import Enum

import re

class Strand(Enum):
    """ File schema independent method to encode alignment strand. """
    FORWARD = 1
    UNKNOWN = 0
    REVERSE = -1

def rec_dd():
    """default_factory for infinitely-recursive defaultdicts.

    Credit
    ======

    https://stackoverflow.com/questions/19189274/defaultdict-of-defaultdict-nested

    :return: infinitely-recursing defaultdict
    :rtype: collections.defaultdict
    """
    return defaultdict(rec_dd)


def merge_intervals(intervals):
    """Generator that yields overlapping intervals from a list.

    Credit
    ======

    https://codereview.stackexchange.com/questions/69242/merging-overlapping-intervals

    :param intervals: a list of tuples (x, y) representing intervals of [x, y]
    :type intervals: list[tuple(numbers.Real, numbers.Real)]
    :return: merged intervals
    :rtype: tuple(numbers.Real, numbers.Real)
    """

    sorted_by_lower_bound = sorted(intervals, key=lambda interval: interval[0])

    if not sorted_by_lower_bound:
        return

    # low and high represent the bounds of the current run of merges
    low, high = sorted_by_lower_bound.pop(0)

    for next_low, next_high in sorted_by_lower_bound:
        if next_low <= high:  # new interval overlaps current run
            high = max(high, next_high)  # merge with the current run
        else:  # current run is over
            yield low, high  # yield accumulated interval
            low, high = next_low, next_high  # start new run

    yield low, high


class Link:
    """
    Represents a connection between two link Features.
    """

    def __init__(self, identifier, target, source, identity):
        """

        :param identifier: unique identifier (within AliTV object) of the Link
        :type identifier: str
        :param target: the connected Feature on the target Chromosome
        :type target: Feature
        :param source: the connected Feature on the source Chromosome
        :type source: Feature
        :param identity: the percent identity of the sequence between the Features linked, multiplied by 100
        :type identity: float
        """

        self.id = identifier
        self.target = target
        self.source = source

        if (identity > 100.0) or (identity < 0.0):
            raise ValueError(
                "'{}' is not a valid identity value for a Link object (0 <= identity <= 100)".format(identity))
        self.identity = identity


class Feature:
    """
    Represents a region on a Chromosome.
    """

    def __init__(self, identifier, name, feature_type, start, end, chromosome):
        """

        :param identifier: unique identifier (within AliTV object) of the Feature
        :type identifier: str
        :param name: the name of the Feature
        :type name: str
        :param feature_type: the type/classification of the feature
        :type feature_type: str
        :param start: the starting base of the feature
        :type start: int
        :param end: the ending base of the feature
        :type end: int
        :param chromosome: the Chromosome containing this Feature
        :type chromosome: Chromosome
        """
        self.id = identifier
        self.name = name
        self.type = feature_type
        self.start = start
        self.end = end
        self.chromosome = chromosome

    def __len__(self):
        return (self.end - self.start) + 1


class Chromosome:
    """
    Represents a contiguous sequence drawn as one box in AliTV.
    """

    def __init__(self, genome, record):
        """

        :param genome: the Genome that this Chromosome belongs to
        :type genome: Genome
        :param record: the Bio.SeqRecord object containing the sequence information of the chromosome
        :type record: Bio.SeqRecord.SeqRecord
        """
        self.genome = genome
        self.seq = record.seq
        self.name = record.id
        self.id = "{}_{}".format(genome.id, self.name)

    def __len__(self):
        return len(self.seq)


class Genome:
    """
    Represents a group of Chromosome objects in a specific order.
    """

    def __init__(self, fasta_file, is_reference=False):
        """

        :param fasta_file: the FASTAFile object containing the Chromosome sequences
        :type fasta_file: minisyn.seq.FASTAFile
        """
        self.fasta = fasta_file
        self.id = fasta_file.name
        self.chromosomes = []
        self._chromosomes_dict = {}
        self.is_reference = is_reference

        for record in fasta_file.records:
            new_chromosome = Chromosome(self, record)
            self.chromosomes.append(new_chromosome)
            self._chromosomes_dict[new_chromosome.name] = new_chromosome

    def __iter__(self):
        return iter(self.chromosomes)

    def get_chromosome_names(self):
        return list(self._chromosomes_dict.keys())

    def get_chromosome_by_name(self, chromosome_name):
        return self._chromosomes_dict[chromosome_name]


class AliTV:
    """
    Represents the AliTV application.
    """

    def __init__(self, fasta_files, minID, maxID):
        """Loads, processes, and converts various input data into the AliTV JSON format.

        :param fasta_files: list of FASTA files
        :type fasta_files: list[minitv.seq.FASTAFile]
        """

        self.links = []
        self.features = []

        self.fasta_files = fasta_files

        self._jsonDict = {
            'data': rec_dd(),
            'conf': rec_dd(),
            'filters': rec_dd()
        }

        # populate karyo/chromosomes
        self.genomes = []
        self._jsonDict['filters']['karyo']['genome_order'] = []
        self._jsonDict['filters']['karyo']['order'] = []

        self.genomes.append(Genome(fasta_files[0], is_reference=True))
        for fasta_file in fasta_files[1:]:
            genome = Genome(fasta_file)
            self.genomes.append(genome)

        for genome in self.genomes:
            for chromosome in genome:
                self._jsonDict['data']['karyo']['chromosomes'][chromosome.id] = {
                    'genome_id': genome.id,
                    'name': chromosome.id,
                    'length': len(chromosome),
                    'seq': None
                }
                self._jsonDict['filters']['karyo']['chromosomes'][chromosome.id] = {
                    'visible': True,
                    'reverse': False
                }

                self._jsonDict['filters']['karyo']['order'].append(chromosome.id)

        self._jsonDict['data']['tree'] = None
        self._jsonDict['conf']['tree']['drawTree'] = False
        self._jsonDict['filters']['karyo']['genome_order'] = [genome.id for genome in self.genomes]

        # default configuration and filters
        self._jsonDict['conf']['linear']['drawAllLinks'] = False
        self._jsonDict['conf']['linear']['startLineColor'] = '#1d91c0'
        self._jsonDict['conf']['linear']['endLineColor'] = '#1d91c0'

        self._jsonDict['conf']['graphicalParameters']['canvasWidth'] = 900
        self._jsonDict['conf']['graphicalParameters']['canvasHeight'] = 900
        self._jsonDict['conf']['graphicalParameters']['karyoHeight'] = 30
        self._jsonDict['conf']['graphicalParameters']['karyoDistance'] = 2000
        self._jsonDict['conf']['graphicalParameters']['linkKaryoDistance'] = 20
        self._jsonDict['conf']['graphicalParameters']['tickLabelFrequency'] = 10
        self._jsonDict['conf']['graphicalParameters']['tickDistance'] = 10000
        self._jsonDict['conf']['graphicalParameters']['treeWidth'] = 200
        self._jsonDict['conf']['graphicalParameters']['genomeLabelWidth'] = 200

        self._jsonDict['conf']['minLinkIdentity'] = minID
        self._jsonDict['conf']['midLinkIdentity'] = (minID+maxID)/2
        self._jsonDict['conf']['maxLinkIdentity'] = maxID

        self._jsonDict['conf']['minLinkIdentityColor'] = '#D21414'
        self._jsonDict['conf']['midLinkIdentityColor'] = '#FFEE05'
        self._jsonDict['conf']['maxLinkIdentityColor'] = '#1DAD0A'
        self._jsonDict['conf']['minLinkLength'] = 100
        self._jsonDict['conf']['maxLinkLength'] = 5000
        self._jsonDict['conf']['layout'] = 'linear'
        self._jsonDict['conf']['features']['showAllFeatures'] = False
        self._jsonDict['conf']['features']['fallbackStyle'] = {
            'form': 'rect',
            'color': '#787878',
            'height': 30,
            'visible': False
        }

        self._jsonDict['conf']['labels']['ticks']['showTicks'] = True
        self._jsonDict['conf']['labels']['ticks']['showTickLabels'] = True
        self._jsonDict['conf']['labels']['chromosome']['showChromosomeLabels'] = False
        self._jsonDict['conf']['labels']['genome']['showGenomeLabels'] = True
        self._jsonDict['conf']['labels']['features']['showFeatureLabels'] = False

        self._jsonDict['filters']['links']['minLinkLength'] = 0
        self._jsonDict['filters']['links']['minLinkIdentity'] = 0
        self._jsonDict['filters']['links']['maxLinkIdentity'] = 100
        self._jsonDict['filters']['onlyShowAdjacentLinks'] = True
        self._jsonDict['filters']['showAllChromosomes'] = False


    def changeColors(self, c):
        o = ["minLinkIdentityColor", "midLinkIdentityColor", "maxLinkIdentityColor"]
        for index, x in enumerate(c):
            if re.search(r'(?:[0-9a-fA-F]{3}){1,2}$', x) and len(x) == 6:
                self._jsonDict["conf"][o[index]] = "#" + x;
            else:
                print("#{} The hex is not working".format(x))

    def optimize_configuration(self):
        """Attempts to optimize AliTV drawing configuration to best display the data"""

        return None

    def set_soft_filters(self, minimum_link_identity, minimum_link_length):
        """Sets soft link filters for AliTV.

        :param minimum_link_identity: minimum link identity for initial AliTV display [0, 100]
        :type minimum_link_identity: float
        :param minimum_link_length: minimum link length for initial AliTV display [0, +inf]
        :type minimum_link_length: float
        """

        if 0 <= minimum_link_identity <= 100:
            self._jsonDict['filters']['links']['minLinkIdentity'] = minimum_link_identity
        else:
            raise ValueError("AliTV minimum link identity must be between 0 and 100")
        if 0 <= minimum_link_length:
            self._jsonDict['filters']['links']['minLinkLength'] = minimum_link_length
        else:
            raise ValueError("AliTV minimum link length must be at least 0")

    def hide_chromosomes_by_alignment_coverage(self, minimum_alignment_coverage):
        """Sets chromosomes invisible based on minimum percent coverage by links.

        :param minimum_alignment_coverage: minimum percent coverage by alignments/links to be visible [0, 100]
        :type minimum_alignment_coverage: float
        """

        alignment_intervals = {}
        for link in self.links:
            if link.target.chromosome not in alignment_intervals:
                alignment_intervals[link.target.chromosome] = []
            alignment_intervals[link.target.chromosome].append((link.target.start, link.target.end))
            if link.source.chromosome not in alignment_intervals:
                alignment_intervals[link.source.chromosome] = []
            alignment_intervals[link.source.chromosome].append((link.source.start, link.source.end))

        self.hide_chromosomes_by_region_coverage(alignment_intervals, minimum_alignment_coverage)

    def hide_chromosomes_by_reference_coverage(self, minimum_reference_coverage):
        """Sets chromosomes invisible based on minimum percent coverage by links to the reference.

        :param minimum_reference_coverage: minimum percent coverage by links to the reference to be visible [0, 100]
        :type minimum_reference_coverage: float
        """

        reference_link_intervals = {}
        for link in self.links:
            if link.target.chromosome.genome.is_reference:
                feature = link.source
                chromosome = feature.chromosome
            elif link.source.chromosome.genome.is_reference:
                feature = link.target
                chromosome = feature.chromosome
            else:
                continue
            if chromosome not in reference_link_intervals:
                reference_link_intervals[chromosome] = []
            reference_link_intervals[chromosome].append((feature.start, feature.end))

        self.hide_chromosomes_by_region_coverage(reference_link_intervals, minimum_reference_coverage)

    def hide_chromosomes_by_region_coverage(self, regions, minimum_region_coverage):
        """Sets chromosomes invisible based on minimum percent coverage by arbitrary regions.

        :param regions: dictionary with Chromosome objects as keys and values of lists of intervals that
            represent regions (overlaps are fine) on the respective Chromosome objects
        :type regions: dict[Chromosome, list[tuple(numbers.Real, numbers.Real)]
        :param minimum_region_coverage: minimum percent coverage by regions to be visible [0, 100]
        """

        for chromosome in regions:
            merged_intervals = merge_intervals(regions[chromosome])
            total_reference_link_coverage = 0
            for low, high in merged_intervals:
                total_reference_link_coverage += (high - low + 1)
            percent_region_coverage = (total_reference_link_coverage / len(chromosome.seq)) * 100
            if percent_region_coverage < minimum_region_coverage:
                print("Hiding chromosome '{}' because its reference coverage is too low ({} < {})".format(
                    chromosome.id, round(percent_region_coverage, 2), minimum_region_coverage
                ))
                self._jsonDict['filters']['karyo']['chromosomes'][chromosome.id]['visible'] = False

    def load_links(self, alignment_groups):
        """

        :param alignment_groups: a dictionary with keys of tuples (target, query) where target and query are names of
          the FASTA files of the target and query of the alignment, respectively
        :type alignment_groups: dict
        """

        # populate links



        ## REDO THIS
        ##TODO
        
        #genome_pairs = tuple(itertools.combinations(self.genomes, r=2))


        genome_pairs = []
        tar = None
        que = None

        for key, value in alignment_groups.items():

            for x in self.genomes:
                if key[0] == x.id:
                    tar = x
                if key[1] == x.id:
                    que = x
            genome_pairs.append((tar, que))

        genome_pairs = tuple(genome_pairs)
        for genome_pair in genome_pairs:
            target_genome = genome_pair[0]
            query_genome = genome_pair[1]
            alignment_group = alignment_groups[target_genome.fasta.name, query_genome.fasta.name]
            for alignment in alignment_group:
                target_chromosome = target_genome.get_chromosome_by_name(alignment.target_name)
                query_chromosome = query_genome.get_chromosome_by_name(alignment.query_name)

                query_block_id = "b{}".format(str(len(self.features)))
                query_block = Feature(query_block_id, '', 'link', alignment.query_start, alignment.query_end,
                                      query_chromosome)
                self.features.append(query_block)

                target_block_id = "b{}".format(str(len(self.features)))
                target_block = Feature(target_block_id, '', 'link', alignment.target_start, alignment.target_end,
                                       target_chromosome)
                self.features.append(target_block)

                # orient query link block according to strand
                if alignment.strand == Strand.REVERSE:
                    self._jsonDict['data']['features']['link'][query_block.id] = {
                        'karyo': query_block.chromosome.id,
                        'start': query_block.end,
                        'end': query_block.start
                    }
                else:
                    self._jsonDict['data']['features']['link'][query_block.id] = {
                        'karyo': query_block.chromosome.id,
                        'start': query_block.start,
                        'end': query_block.end
                    }

                self._jsonDict['data']['features']['link'][target_block.id] = {
                    'karyo': target_block.chromosome.id,
                    'start': target_block.start,
                    'end': target_block.end
                }

                # AliTV only needs one-way links
                link_id = "link{}".format(len(self.links))
                link = Link(link_id, target_block, query_block, round(alignment.identity, 2))
                self.links.append(link)

                self._jsonDict['data']['links'][query_genome.id][target_genome.id][link_id] = {
                    'target': target_block.id,
                    'source': query_block.id,
                    'identity': link.identity
                }

    def order_and_orient_sequences(self, alignment_groups):
        """Orders and orients each Chromosome in non-reference Genomes based on the first longest alignment between each
        Chromosome and a reference Chromosome.

        The "reference" genome is currently the first genome in the list loaded into AliTV.

        :param alignment_groups: a dictionary with keys of tuples (target, query) where target and query are names of
          the FASTA files of the target and query of the alignment, respectively
        :type alignment_groups: dict
        """

        self._jsonDict['filters']['karyo']['order'] = []

        reference_genome = self.genomes[0]
        for chromosome in reference_genome:
            self._jsonDict['filters']['karyo']['order'].append(chromosome.id)

        for genome in self.genomes[1:]:
            try:
                alignment_group = alignment_groups[reference_genome.fasta.name, genome.fasta.name]
                reference_attr_prefix = 'target'
                non_reference_attr_prefix = 'query'
            except KeyError:
                alignment_group = alignment_groups[genome.fasta.name, reference_genome.fasta.name]
                reference_attr_prefix = 'query'
                non_reference_attr_prefix = 'target'

            longest_alignments = defaultdict(list)
            for alignment in alignment_group:

                # find the longest alignment for each non-reference chromosome
                non_reference_chromosome_name = alignment.__getattribute__(non_reference_attr_prefix + "_name")
                if len(longest_alignments[non_reference_chromosome_name]) < len(alignment):
                    longest_alignments[non_reference_chromosome_name] = alignment

            aligned_chromosomes = set(longest_alignments.keys())
            all_chromosome_names = set([chromosome.name for chromosome in genome])
            reversed_chromosomes = set()
            reference_query_chromosome_order = {}
            for chromosome in reference_genome:
                reference_query_chromosome_order[chromosome.name] = []

            for aligned_chromosome in aligned_chromosomes:
                longest_alignment = longest_alignments[aligned_chromosome]

                if longest_alignment.strand == '-':
                    reversed_chromosomes.add(aligned_chromosome)

                reference_chromosome_name = longest_alignment.__getattribute__(reference_attr_prefix + "_name")
                reference_chromosome_start = longest_alignment.__getattribute__(reference_attr_prefix + "_start")
                non_reference_chromosome_name = longest_alignment.__getattribute__(non_reference_attr_prefix + "_name")
                reference_query_chromosome_order[reference_chromosome_name].append(
                    (genome.get_chromosome_by_name(non_reference_chromosome_name).id,
                     reference_chromosome_start)
                )

            for reversed_chromosome in reversed_chromosomes:
                alitv_karyo_id = genome.get_chromosome_by_name(reversed_chromosome).id
                self._jsonDict['filters']['karyo']['chromosomes'][alitv_karyo_id]['reverse'] = True

            for reference_chromosome_name in reference_query_chromosome_order:
                sorted_query_order = sorted(reference_query_chromosome_order[reference_chromosome_name],
                                            key=lambda x: x[1])
                sorted_query_order_names = [pair[0] for pair in sorted_query_order]
                self._jsonDict['filters']['karyo']['order'] += sorted_query_order_names
            unaligned_chromosome_names = list(all_chromosome_names - aligned_chromosomes)
            unaligned_chromosome_ids = [genome.get_chromosome_by_name(chromosome_name).id for chromosome_name in
                                        unaligned_chromosome_names]
            self._jsonDict['filters']['karyo']['order'] += list(unaligned_chromosome_ids)

    def load_annotations(self, annotations, annotation_types):
        """

        :param annotations: iterable of minitv.annotation.GFF3Annotation objects from genome GFF3 files
        :type annotations: collections.Iterable[minitv.annotation.GFF3Annotation]
        :param annotation_types: iterable of strings of feature types to filter the annotation
        :type annotation_types: collections.Iterable[str]
        """

        for i, annotation in enumerate(annotations):
            genome = self.genomes[i]

            for feature in annotation.features:
                if feature.type not in self._jsonDict['data']['features']:
                    self._jsonDict['data']['features'][feature.type] = []

                try:
                    chromosome = genome.get_chromosome_by_name(feature.seqid)
                    new_feature = {
                        'karyo': chromosome.id,
                        'start': feature.start,
                        'end': feature.end
                    }
                    if hasattr(feature, 'name'):
                        new_feature['name'] = feature.name
                    self._jsonDict['data']['features'][feature.type].append(new_feature)
                except KeyError:
                    # Feature is on a chromosome not loaded into AliTV, skip
                    pass

        for annotation_type in annotation_types:
            self._jsonDict['conf']['features']['supportedFeatures'][annotation_type] = {
                'form': 'rect',
                'color': "#E2EDFF",
                'height': 30,
                'visible': True
            }



    def get_json(self, indent):
        """

        :return: AliTV JSON data
        :rtype: str
        """

        return json.dumps(self._jsonDict, indent=indent)


from setuptools import setup, find_packages

setup(
    name="paftv10",
    version='0.1.1',
    packages= find_packages(),
    install_requires=[
          'biopython',
      ],
    entry_points={
        'console_scripts': [
            'paftv = paftv10.paftv:main',
        ],
    },
    license='MIT',
    description='Visualization for PAF files using AliTV',
    author='Sebastian Vorbrugg',
    author_email='sebastian.vorbrugg@tuebingen.mpg.de',
    url='https://github.com/MoinSebi/paftv'

)
